#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>

#include <gst/gst.h>

#ifdef USE_GL
#  include "glad/glad.h"
#else
#  include "glad/glad_es.h"
#endif
#include "glad/glad_egl.h"

#ifndef FALSE
#  define FALSE 0
#endif
#ifndef TRUE
#  define TRUE 1
#endif

#define IMAGE_WIDTH  640
#define IMAGE_HEIGHT 480

const char* eglErrorString(EGLint error)
{
    switch(error)
    {
    case EGL_SUCCESS: return "No error";
    case EGL_NOT_INITIALIZED: return "EGL not initialized or failed to initialize";
    case EGL_BAD_ACCESS: return "Resource inaccessible";
    case EGL_BAD_ALLOC: return "Cannot allocate resources";
    case EGL_BAD_ATTRIBUTE: return "Unrecognized attribute or attribute value";
    case EGL_BAD_CONTEXT: return "Invalid EGL context";
    case EGL_BAD_CONFIG: return "Invalid EGL frame buffer configuration";
    case EGL_BAD_CURRENT_SURFACE: return "Current surface is no longer valid";
    case EGL_BAD_DISPLAY: return "Invalid EGL display";
    case EGL_BAD_SURFACE: return "Invalid surface";
    case EGL_BAD_MATCH: return "Inconsistent arguments";
    case EGL_BAD_PARAMETER: return "Invalid argument";
    case EGL_BAD_NATIVE_PIXMAP: return "Invalid native pixmap";
    case EGL_BAD_NATIVE_WINDOW: return "Invalid native window";
    case EGL_CONTEXT_LOST: return "Context lost";
    }
    return "Unknown error ";
}

const char* eglStreamStateString(EGLint streamState) {
    switch(streamState) {
        case EGL_STREAM_STATE_CREATED_KHR: return "Stream created";
        case EGL_STREAM_STATE_CONNECTING_KHR: return "Consumer connected, producer not connected";
        case EGL_STREAM_STATE_EMPTY_KHR: return "Consumer and producer connected, waiting for first frame";
        case EGL_STREAM_STATE_NEW_FRAME_AVAILABLE_KHR: return "New frame available";
        case EGL_STREAM_STATE_OLD_FRAME_AVAILABLE_KHR: return "Old frame available";
        case EGL_STREAM_STATE_DISCONNECTED_KHR: return "Disconnected state";
    }
    return "Unknown state ";
}

#define egl_assert(stmt) \
    { \
        EGLBoolean res = stmt; \
        if ( res == EGL_FALSE ) {\
            printf("EGL ERROR in file %s @ line %d\n",  __FILE__, __LINE__);}\
        EGLint eglErr; \
        eglErr = eglGetError(); \
        if (eglErr != EGL_SUCCESS) \
        { \
            printf("glError in file %s @ line %d: %x %s\n", \
                     __FILE__, __LINE__, eglErr, eglErrorString(eglErr)); \
            abort();\
        } \
    } (void)0

void printStreamState(EGLDisplay display, EGLStreamKHR stream) {
    EGLint streamState;
    egl_assert(eglQueryStreamKHR(display, stream, EGL_STREAM_STATE_KHR, &streamState));
    printf("%s\n", eglStreamStateString(streamState));
}

#define esLogMessage(...) fprintf(stderr, __VA_ARGS__);

typedef struct ESContext ESContext;

struct ESContext {
    void *userData;

    GLint width;
    GLint height;

    EGLDisplay   display;
    EGLConfig    config;
    EGLContext   context;
    EGLSurface   surface;
    EGLStreamKHR stream;

    GstElement *pipeline;
};

static ESContext esContext;

EGLint GetContextRenderableType(EGLDisplay display) {
#ifdef EGL_KHR_create_context
    const char *extensions = eglQueryString(display, EGL_EXTENSIONS);

    if (extensions != NULL && strstr(extensions, "EGL_KHR_create_context")) {
        return EGL_OPENGL_ES3_BIT_KHR;
    }
#endif
    return EGL_OPENGL_ES2_BIT;
}

typedef struct {
    GLuint programObject;
} UserData;

///
// Create a shader object, load the shader source, and
// compile the shader.
//
GLuint LoadShader ( GLenum type, const char *shaderSrc )
{
   GLuint shader;
   GLint compiled;

   // Create the shader object
   shader = glCreateShader ( type );

   if ( shader == 0 )
   {
      return 0;
   }

   // Load the shader source
   glShaderSource ( shader, 1, &shaderSrc, NULL );

   // Compile the shader
   glCompileShader ( shader );

   // Check the compile status
   glGetShaderiv ( shader, GL_COMPILE_STATUS, &compiled );

   if ( !compiled )
   {
      GLint infoLen = 0;

      glGetShaderiv ( shader, GL_INFO_LOG_LENGTH, &infoLen );

      if ( infoLen > 1 )
      {
         char *infoLog = malloc ( sizeof ( char ) * infoLen );

         glGetShaderInfoLog ( shader, infoLen, NULL, infoLog );
         esLogMessage ( "Error compiling shader:\n%s\n", infoLog );

         free ( infoLog );
      }

      glDeleteShader ( shader );
      return 0;
   }

   return shader;

}

int Init() {
    UserData *userData = esContext.userData;
    char vShaderStr[] =
       "#version 300 es                          \n"
       "layout(location = 0) in vec4 vPosition;  \n"
       "void main()                              \n"
       "{                                        \n"
       "   gl_Position = vPosition;              \n"
       "}                                        \n";

    char fShaderStr[] =
       "#version 300 es                              \n"
       "precision mediump float;                     \n"
       "out vec4 fragColor;                          \n"
       "void main()                                  \n"
       "{                                            \n"
       "   fragColor = vec4 ( 1.0, 0.0, 0.0, 1.0 );  \n"
       "}                                            \n";

    GLuint vertexShader;
    GLuint fragmentShader;
    GLuint programObject;
    GLint linked;

    // Load the vertex/fragment shaders
    vertexShader = LoadShader ( GL_VERTEX_SHADER, vShaderStr );
    fragmentShader = LoadShader ( GL_FRAGMENT_SHADER, fShaderStr );

    // Create the program object
    programObject = glCreateProgram ( );

    if ( programObject == 0 )
    {
       return 0;
    }

    glAttachShader ( programObject, vertexShader );
    glAttachShader ( programObject, fragmentShader );

    // Link the program
    glLinkProgram ( programObject );

    // Check the link status
    glGetProgramiv ( programObject, GL_LINK_STATUS, &linked );

    if ( !linked )
    {
       GLint infoLen = 0;

       glGetProgramiv ( programObject, GL_INFO_LOG_LENGTH, &infoLen );

       if ( infoLen > 1 )
       {
          char *infoLog = malloc ( sizeof ( char ) * infoLen );

          glGetProgramInfoLog ( programObject, infoLen, NULL, infoLog );
          esLogMessage ( "Error linking program:\n%s\n", infoLog );

          free ( infoLog );
       }

       glDeleteProgram ( programObject );
       return FALSE;
    }

     // Store the program object
     userData->programObject = programObject;

     // Delete shaders
     glDeleteShader(vertexShader);
     glDeleteShader(fragmentShader);

     glClearColor ( 1.0f, 1.0f, 1.0f, 0.0f );

     return TRUE;
}

float color = 0.0f;

///
// Draw a triangle using the shader pair created in Init()
//
void Draw ()
{
    UserData *userData = esContext.userData;

   GLfloat vVertices[] = {  0.0f,  0.5f, 0.0f,
                            -0.5f, -0.5f, 0.0f,
                            0.5f, -0.5f, 0.0f
                         };

   // Set the viewport
   glViewport ( 0, 0, esContext.width, esContext.height );

   color = fmod(color + 0.01f, 1.0f);
   glClearColor(color, color, color, 0.0f);

   // Clear the color buffer
   glClear ( GL_COLOR_BUFFER_BIT );

   // Use the program object
   glUseProgram ( userData->programObject );

   // Load the vertex data
   glVertexAttribPointer ( 0, 3, GL_FLOAT, GL_FALSE, 0, vVertices );
   glEnableVertexAttribArray ( 0 );

   glDrawArrays ( GL_TRIANGLES, 0, 3 );
}

void Shutdown () {
    UserData *userData = esContext.userData;

    glDeleteProgram(userData->programObject);
}

int InitGST() {
    GError *err = NULL;

    if (!gst_init_check(NULL, NULL, &err)) {
        g_print("Failed to init gstreamer: %s\n", err->message);
        g_error_free(err);
        return FALSE;
    }

    char*  launchString = NULL;
    size_t bufferSize   = 0;
    FILE*  myStream     = open_memstream(&launchString, &bufferSize);

    fprintf(myStream, "nveglstreamsrc name=egl_src ! ");
    fprintf(myStream, "video/x-raw(memory:NVMM), format=I420, width=%i, height=%i, framerate=30/1 ! ", esContext.width, esContext.height);
    fprintf(myStream, "omxh264enc bitrate=2500000 name=video_enc iframeinterval=15 profile=high control-rate=4 ! queue ! ");
    fprintf(myStream, "matroskamux ! ");
#ifdef USE_GL
    fprintf(myStream, "filesink location=test_gl.mkv");
#else
    fprintf(myStream, "filesink location=test_gles.mkv");
#endif

    fclose(myStream);

    esContext.pipeline = gst_parse_launch(launchString, &err);
    free(launchString);

    if (esContext.pipeline == NULL) {
        g_print("Failed to parse launch: %s\n", err->message);
        g_error_free(err);
        return FALSE;
    }

    GstElement *eglStreamSrc = gst_bin_get_by_name(GST_BIN(esContext.pipeline), "egl_src");
    if (!GST_IS_ELEMENT(eglStreamSrc)) {
        g_print("Failed to get source from pipeline \n");
        return FALSE;
    }

    g_object_set(eglStreamSrc,
        "display",   esContext.display,
        "eglstream", esContext.stream,
        NULL);

    gst_element_set_state(esContext.pipeline, GST_STATE_PLAYING);

    return TRUE;
}

void ShutdownGST() {
    gst_element_set_state(esContext.pipeline, GST_STATE_NULL);
    gst_element_get_state(esContext.pipeline, NULL, NULL, GST_CLOCK_TIME_NONE);

    gst_object_unref(GST_OBJECT(esContext.pipeline));
    gst_deinit();
}

int esMain() {
    EGLint majorVersion;
    EGLint minorVersion;
#ifdef USE_GL
    EGLint contextAttribs[] = { EGL_NONE };
#else
    EGLint contextAttribs[] = {EGL_CONTEXT_CLIENT_VERSION, 3, EGL_NONE };
#endif

    esContext.width  = IMAGE_WIDTH;
    esContext.height = IMAGE_HEIGHT;

    // Initialize GLAD EGL
    int res = gladLoadEGL();
    if (res < 0) {
        return GL_FALSE;
    }

    esContext.display = eglGetDisplay(NULL);
    if (esContext.display == EGL_NO_DISPLAY) {
        return GL_FALSE;
    }

    if (!eglInitialize(esContext.display, &majorVersion, &minorVersion)) {
        return GL_FALSE;
    }

    printf("Loaded EGL Version: %i.%i\n", majorVersion, minorVersion);

    {
        EGLint numConfigs = 0;
        EGLint attribList[] = {
            EGL_RED_SIZE,        8,
            EGL_GREEN_SIZE,      8,
            EGL_BLUE_SIZE,       8,
            EGL_ALPHA_SIZE,      EGL_DONT_CARE,
            EGL_DEPTH_SIZE,      EGL_DONT_CARE,
            EGL_STENCIL_SIZE,    EGL_DONT_CARE,
            EGL_SAMPLE_BUFFERS,  1,
#ifdef USE_GL
            EGL_RENDERABLE_TYPE, EGL_OPENGL_BIT,
#else
            EGL_RENDERABLE_TYPE, GetContextRenderableType(esContext.display),
#endif
            EGL_NONE
        };

        if (!eglChooseConfig(esContext.display, attribList, &esContext.config, 1, &numConfigs)) {
            return GL_FALSE;
        }

        if (numConfigs < 1) {
            return GL_FALSE;
        }
    }

#ifdef USE_GL
    if (!eglBindAPI(EGL_OPENGL_API)) {
        return GL_FALSE;
    }
#endif

    EGLint streamAttrMailboxMode[] = {
        EGL_NONE
    };

    esContext.stream = eglCreateStreamKHR(esContext.display, streamAttrMailboxMode);

    printStreamState(esContext.display, esContext.stream);

    esContext.context = eglCreateContext(esContext.display, esContext.config, EGL_NO_CONTEXT, contextAttribs);
    if (esContext.context == EGL_NO_CONTEXT) {
        return GL_FALSE;
    }

    if (!eglMakeCurrent(esContext.display, EGL_NO_SURFACE,
                        EGL_NO_SURFACE, esContext.context)) {
        return GL_FALSE;
    }

#ifdef USE_GL
    res = gladLoadGLLoader((GLADloadproc) eglGetProcAddress);
    if (res < 0) {
        return GL_FALSE;
    }
    printf("Loaded OpenGL Version: %i.%i\n", GLVersion.major, GLVersion.minor);
#else
    res = gladLoadGLES2Loader((GLADloadproc) eglGetProcAddress);
    if (res < 0) {
        return GL_FALSE;
    }
    printf("Loaded OpenGL ES Version: %i.%i\n", GLVersion.major, GLVersion.minor);
#endif

    if ( !Init() ) {
        return GL_FALSE;
    }

    if ( !InitGST() ) {
        return GL_FALSE;
    }

    printStreamState(esContext.display, esContext.stream);

    EGLint attrs[] = {
        EGL_WIDTH,  esContext.width,
        EGL_HEIGHT, esContext.height,
        EGL_NONE
    };

    if (!eglMakeCurrent(esContext.display, EGL_NO_SURFACE,
                        EGL_NO_SURFACE, EGL_NO_CONTEXT)) {
        return GL_FALSE;
    }

    esContext.surface = eglCreateStreamProducerSurfaceKHR(esContext.display, esContext.config, esContext.stream, attrs);
    if (esContext.surface == EGL_NO_SURFACE) {
        return GL_FALSE;
    }

    if (!eglMakeCurrent(esContext.display, EGL_NO_SURFACE,
                        EGL_NO_SURFACE, esContext.context)) {
        return GL_FALSE;
    }

    printStreamState(esContext.display, esContext.stream);

    return GL_TRUE;
}

int main(int argc, char *argv[]) {
    memset(&esContext, 0, sizeof(esContext));
    esContext.userData = malloc(sizeof(UserData));

    if (esMain() != GL_TRUE)
        return -1;

    egl_assert(eglMakeCurrent(esContext.display, esContext.surface,
                              esContext.surface, esContext.context));

    int i;
    for (i = 0; i < 600; i++) {
        // printStreamState(esContext.display, esContext.stream);

        Draw();

        egl_assert(eglSwapBuffers(esContext.display, esContext.surface));

        // printStreamState(esContext.display, esContext.stream);

        usleep(1000000 / 30);
    }

    usleep(1000000);

    ShutdownGST();
    Shutdown();

    if (esContext.userData != NULL)
        free (esContext.userData);

    return 0;
}
